from qiskit.providers import JobStatus

from qctic.schemas import GetJobsQuerySchema


def test_get_jobs_query():
    limit = 15

    jobs_query = GetJobsQuerySchema().dump({
        "limit": limit
    })

    assert jobs_query["limit"] is limit
    assert jobs_query.get("status", None) is None

    status_str = JobStatus.RUNNING.name

    jobs_query = GetJobsQuerySchema().dump({
        "limit": limit,
        "status": status_str
    })

    assert jobs_query["limit"] is limit
    assert jobs_query["status"] == [status_str]

    status_list = [
        JobStatus.RUNNING.name,
        JobStatus.DONE.name
    ]

    jobs_query = GetJobsQuerySchema().dump({
        "limit": limit,
        "status": status_list
    })

    assert jobs_query["limit"] is limit
    assert jobs_query["status"] == status_list



import json
import time
from datetime import datetime, timedelta, timezone

import pytest
from qiskit import Aer, execute
from qiskit.assembler import disassemble
from qiskit.providers import JobStatus
from werkzeug.wrappers import Response

from qctic.utils import QcticResultTimeoutError, wait_result, wait_result_async

_RESULT_SLEEP = 0.25


def _mock_simulator_result(job, httpserver, latency=_RESULT_SLEEP):
    circuits, _run_config, _user_qobj_header = disassemble(job.qobj())
    simulator = Aer.get_backend("qasm_simulator")
    simulator_job = execute(circuits, simulator)
    simulator_result = simulator_job.result()
    url = "/jobs/{}".format(job.job_id())
    dtime_now = datetime.now(timezone.utc)
    dtime_end = dtime_now + timedelta(seconds=latency)
    tstamp_end = dtime_now.timestamp() + latency

    job_dict = {
        "job_id": job.job_id(),
        "qobj": job.qobj().to_dict(),
        "date_submit": dtime_now.isoformat(),
        "date_start": dtime_now.isoformat(),
        "status": JobStatus.RUNNING.name
    }

    def handler(req):
        job_res = {**job_dict}

        if time.time() >= tstamp_end:
            job_res.update({
                "date_end": dtime_end.isoformat(),
                "status": JobStatus.DONE.name,
                "result": simulator_result.to_dict()
            })

        return Response(json.dumps(job_res))

    httpserver.expect_request(url).respond_with_handler(handler)

    return simulator_result


def test_wait_result(job, httpserver):
    simulator_result = _mock_simulator_result(job, httpserver)
    result = wait_result(job, base_sleep=_RESULT_SLEEP)
    assert result.to_dict() == simulator_result.to_dict()


@pytest.mark.asyncio
async def test_wait_result_async(job, httpserver):
    simulator_result = _mock_simulator_result(job, httpserver)
    result = await wait_result_async(job, base_sleep=_RESULT_SLEEP)
    assert result.to_dict() == simulator_result.to_dict()


@pytest.mark.asyncio
async def test_wait_timeout(job, httpserver):
    base_sleep = 0.1
    sleep_growth = 1.0
    timeout = base_sleep
    latency = base_sleep * 20.0

    _mock_simulator_result(job, httpserver, latency=latency)

    with pytest.raises(QcticResultTimeoutError):
        await wait_result_async(
            job,
            base_sleep=base_sleep,
            timeout=timeout,
            sleep_growth=sleep_growth)



import asyncio
import copy
import logging
import uuid

from qiskit.providers import BaseBackend
from qiskit.providers.aer.backends import (QasmSimulator, StatevectorSimulator,
                                           UnitarySimulator)
from qiskit.providers.models import BackendStatus, QasmBackendConfiguration
from qiskit.qobj import QasmQobj as Qobj

from qctic.__version__ import __version__
from qctic.job import QCticJob

_logger = logging.getLogger(__name__)


class BaseQCticSimulator(BaseBackend):
    """Base class for CTIC Erwin simulator backends."""

    N_QUBITS = 32
    MAX_EXPERIMENTS = 20

    BASE_DESCRIPTION = (
        "A remote quantum simulator based on "
        "Aer's {} that runs on the "
        "high-performance CTIC Erwin computer"
    )

    BASE_CONFIGURATION = {
        "backend_version": __version__,
        "n_qubits": N_QUBITS,
        "url": "https://bitbucket.org/fundacionctic/erwin-qiskit",
        "local": False,
        "max_experiments": MAX_EXPERIMENTS
    }

    def __init__(self, *args, **kwargs):
        """Constructor."""

        kwargs["configuration"] = self.default_configuration()
        super().__init__(*args, **kwargs)

    @classmethod
    def default_configuration(cls):
        """dict: Default base configuration for this backend class."""

        raise NotImplementedError

    @property
    def api(self):
        """QCticAPI: The API instance of this backend's provider."""

        return self.provider().api

    def _res_to_job(self, res):
        return QCticJob(self, res["job_id"], Qobj.from_dict(res["qobj"]), remote_job=res)

    def _default_backend_status(self):
        return BackendStatus(
            backend_name=self.name(),
            backend_version=self.configuration().backend_version,
            operational=False,
            pending_jobs=0,
            status_msg="The Erwin quantum simulation platform is not currently operational")

    def _res_to_backend_status(self, res):
        return BackendStatus(
            backend_name=self.name(),
            backend_version=self.configuration().backend_version,
            operational=res.get("operational"),
            pending_jobs=res.get("pending_jobs"),
            status_msg="The Erwin quantum simulation platform is operational")

    def _log_status_ex(self, ex):
        _logger.warning("Error fetching status: {}".format(ex))

    def run(self, qobj, **kwargs):
        """Run a qobj on the backend.

        Args:
            qobj (Qobj): The Qobj to be executed.

        Returns:
            QCticJob: The simulation job.

        Additional Information:
            * The ``kwargs`` will be passed to the ``run`` method of the 
              ``AerBackend`` instance in the remote simulation platform.
        """

        job_id = str(uuid.uuid4())
        job = QCticJob(self, job_id, qobj, run_params=kwargs)

        if not self.api.loop.is_running():
            _logger.debug("Synchronous job submit: %s", job_id)
            job.submit()
        else:
            _logger.debug("Asynchronous job submit: %s", job_id)
            job.submit_task = asyncio.ensure_future(job.submit_async())

        return job

    def status(self):
        """Fetch the current status of the simulation platform.

        Returns:
            BackendStatus: The status.
        """

        try:
            res = self.api.get_backend_status_sync()
            return self._res_to_backend_status(res)
        except Exception as ex:
            self._log_status_ex(ex)
            return self._default_backend_status()

    def jobs(self, **kwargs):
        """Fetch a set of jobs that match the given filters.

        The filters can be passed as optional keyword arguments 
        that match those of the ``QCticAPI.get_jobs`` method.

        Returns:
            list(QCticJob): List of jobs.
        """

        res = self.api.get_jobs_sync(**kwargs)
        return [self._res_to_job(item) for item in res]

    def retrieve_job(self, job_id):
        """Fetch a job given its ID.

        Args:
            job_id (str): Job ID.

        Returns:
            QCticJob: The job.
        """

        job_res = self.api.get_job_sync(job_id)
        return self._res_to_job(job_res)

    async def status_async(self):
        """Asynchronous version of the ``status`` method."""

        try:
            res = await self.api.get_backend_status()
            return self._res_to_backend_status(res)
        except Exception as ex:
            self._log_status_ex(ex)
            return self._default_backend_status()

    async def jobs_async(self, **kwargs):
        """Asynchronous version of the ``jobs`` method."""

        res = await self.api.get_jobs(**kwargs)
        return [self._res_to_job(item) for item in res]

    async def retrieve_job_async(self, job_id):
        """Asynchronous version of the ``retrieve_job`` method."""

        res = await self.api.get_job(job_id)
        return self._res_to_job(res)


class QCticQasmSimulator(BaseQCticSimulator):
    """Backend that represents the CTIC Erwin simulation platform.

    Remote jobs are executed on Aer's QasmSimulator."""

    NAME = "ctic_erwin_qasm_simulator"

    @classmethod
    def default_configuration(cls):
        """Returns the default configuration for the Erwin QasmSimulator.

        Returns:
            QasmBackendConfiguration: The configuration object.
        """

        conf = copy.deepcopy(QasmSimulator.DEFAULT_CONFIGURATION)
        conf.update(cls.BASE_CONFIGURATION)

        conf.update({
            "backend_name": cls.NAME,
            "description": cls.BASE_DESCRIPTION.format("QasmSimulator")
        })

        return QasmBackendConfiguration.from_dict(conf)


class QCticStatevectorSimulator(BaseQCticSimulator):
    """Backend that represents the CTIC Erwin simulation platform.

    Remote jobs are executed on Aer's StatevectorSimulator."""

    NAME = "ctic_erwin_statevector_simulator"

    @classmethod
    def default_configuration(cls):
        """Returns the default configuration for the Erwin StatevectorSimulator.

        Returns:
            QasmBackendConfiguration: The configuration object.
        """

        conf = copy.deepcopy(StatevectorSimulator.DEFAULT_CONFIGURATION)
        conf.update(cls.BASE_CONFIGURATION)

        conf.update({
            "backend_name": cls.NAME,
            "description": cls.BASE_DESCRIPTION.format("StatevectorSimulator")
        })

        return QasmBackendConfiguration.from_dict(conf)


class QCticUnitarySimulator(BaseQCticSimulator):
    """Backend that represents the CTIC Erwin simulation platform.

    Remote jobs are executed on Aer's UnitarySimulator."""

    NAME = "ctic_erwin_unitary_simulator"

    @classmethod
    def default_configuration(cls):
        """Returns the default configuration for the Erwin UnitarySimulator.

        Returns:
            QasmBackendConfiguration: The configuration object.
        """

        conf = copy.deepcopy(UnitarySimulator.DEFAULT_CONFIGURATION)
        conf.update(cls.BASE_CONFIGURATION)

        conf.update({
            "backend_name": cls.NAME,
            "description": cls.BASE_DESCRIPTION.format("UnitarySimulator")
        })

        return QasmBackendConfiguration.from_dict(conf)



import asyncio
import time

_BASE_SLEEP = 1.0
_MAX_SLEEP = 30.0
_SLEEP_GROWTH = 1.2
_TIMEOUT = None


class QcticResultTimeoutError(Exception):
    pass


def _raise_if_timeout(start, timeout):
    if timeout is None:
        return

    diff = time.time() - start

    if diff >= timeout:
        raise QcticResultTimeoutError(
            "Timeout exceeded ({} secs)".format(timeout))


def wait_result(
        job, base_sleep=_BASE_SLEEP, max_sleep=_MAX_SLEEP,
        sleep_growth=_SLEEP_GROWTH, timeout=_TIMEOUT):
    sleep = base_sleep
    start = time.time()

    while True:
        res = job.result(fetch=True)

        if res:
            return res

        _raise_if_timeout(start, timeout)
        time.sleep(sleep)
        sleep = min(sleep * sleep_growth, max_sleep)


async def wait_result_async(
        job, base_sleep=_BASE_SLEEP, max_sleep=_MAX_SLEEP,
        sleep_growth=_SLEEP_GROWTH, timeout=_TIMEOUT):
    sleep = base_sleep
    start = time.time()

    while True:
        res = await job.result_async(fetch=True)

        if res:
            return res

        _raise_if_timeout(start, timeout)
        await asyncio.sleep(sleep)
        sleep = min(sleep * sleep_growth, max_sleep)



import datetime
import json
import os
import random
import uuid

import pytest
from qiskit import Aer, execute
from qiskit.assembler import disassemble
from qiskit.providers import JobStatus
from tests.utils import fake_job_dict, fake_qobj
from werkzeug.wrappers import Response

from qctic.backend import QCticQasmSimulator
from qctic.job import QCticJob
from qctic.provider import QCticProvider

_LOCALHOST = "http://localhost"
_JOB_DICTS = [fake_job_dict() for _ in range(random.randint(1, 10))]
_PENDING_JOBS = random.randint(0, 10)
_JOB_DICT = fake_job_dict()
_STATUS = JobStatus.QUEUED.name


def _clear_api_env():
    env_vars = [
        QCticProvider.ENV_HOST,
        QCticProvider.ENV_USERNAME,
        QCticProvider.ENV_PASSWORD,
        QCticProvider.ENV_TOKEN
    ]

    for item in env_vars:
        os.environ[item] = ""


@pytest.fixture
def clear_env():
    _clear_api_env()
    yield True
    _clear_api_env()


def test_from_env_token(clear_env):
    with pytest.raises(Exception):
        QCticProvider.from_env()

    os.environ[QCticProvider.ENV_HOST] = _LOCALHOST
    os.environ[QCticProvider.ENV_TOKEN] = str(uuid.uuid4())

    assert QCticProvider.from_env()


def test_from_env_basic(clear_env):
    with pytest.raises(Exception):
        QCticProvider.from_env()

    os.environ[QCticProvider.ENV_HOST] = _LOCALHOST
    os.environ[QCticProvider.ENV_USERNAME] = str(uuid.uuid4())

    with pytest.raises(Exception):
        QCticProvider.from_env()

    os.environ[QCticProvider.ENV_PASSWORD] = str(uuid.uuid4())

    assert QCticProvider.from_env()


def test_provider_get_backend(provider):
    all_backends = provider.backends()
    simulator_backends = provider.backends(simulator=True)

    large_backends = provider.backends(
        filters=lambda x: x.configuration().n_qubits >= 10)

    qasm_backend = provider.get_backend(QCticQasmSimulator.NAME)

    assert len(all_backends) > 0
    assert len(simulator_backends) == len(all_backends)
    assert len(large_backends) == len(all_backends)
    assert qasm_backend
    assert qasm_backend in all_backends
    assert qasm_backend in simulator_backends
    assert qasm_backend in large_backends


def test_backend_base_interface(provider):
    backend = provider.get_backend(QCticQasmSimulator.NAME)
    config = backend.configuration()

    assert backend.provider() == provider
    assert backend.name() == QCticQasmSimulator.NAME
    assert backend.status().backend_name == QCticQasmSimulator.NAME
    assert config.backend_name == backend.status().backend_name
    assert config.backend_version
    assert not config.local


def test_backend_run(provider, httpserver):
    arg_key = str(uuid.uuid4())

    def handler(req):
        data = json.loads(req.data)
        assert data["job_id"]
        assert data["run_params"][arg_key]
        return Response()

    httpserver.expect_request("/jobs").respond_with_handler(handler)
    backend = provider.get_backend(QCticQasmSimulator.NAME)
    qobj = fake_qobj()
    job = backend.run(qobj, **{arg_key: True})

    assert job.job_id()
    assert job.api
    assert job.qobj().to_dict() == qobj.to_dict()
    assert job.run_params[arg_key]


@pytest.fixture
def backend_for_status_test(provider, httpserver):
    httpserver.expect_request("/status").respond_with_json({
        "operational": True,
        "pending_jobs": _PENDING_JOBS
    })

    backend = provider.get_backend(QCticQasmSimulator.NAME)

    return backend


def assert_status_test(status):
    assert status.operational
    assert status.pending_jobs == _PENDING_JOBS


def test_backend_status(backend_for_status_test):
    status = backend_for_status_test.status()
    assert_status_test(status)


@pytest.mark.asyncio
async def test_backend_status_async(backend_for_status_test):
    status = await backend_for_status_test.status_async()
    assert_status_test(status)


@pytest.fixture
def backend_for_jobs_test(provider, httpserver):
    def handler(req):
        assert int(req.args["limit"]) == len(_JOB_DICTS)
        return Response(json.dumps(_JOB_DICTS))

    httpserver.expect_request("/jobs").respond_with_handler(handler)
    backend = provider.get_backend(QCticQasmSimulator.NAME)

    return backend


def assert_jobs_test(jobs):
    assert len(jobs) == len(_JOB_DICTS)

    assert all([
        next(True for item in _JOB_DICTS if item["job_id"] == job.job_id())
        for job in jobs
    ])


def test_backend_jobs(backend_for_jobs_test):
    jobs = backend_for_jobs_test.jobs(limit=len(_JOB_DICTS))
    assert_jobs_test(jobs)


@pytest.mark.asyncio
async def test_backend_jobs_async(backend_for_jobs_test):
    jobs = await backend_for_jobs_test.jobs_async(limit=len(_JOB_DICTS))
    assert_jobs_test(jobs)


@pytest.fixture
def backend_for_retrieve_test(provider, httpserver):
    def handler(req):
        return Response(json.dumps(_JOB_DICT))

    url = "/jobs/{}".format(_JOB_DICT["job_id"])
    httpserver.expect_request(url).respond_with_handler(handler)
    backend = provider.get_backend(QCticQasmSimulator.NAME)

    return backend


def test_backend_retrieve_job(backend_for_retrieve_test):
    job_id = _JOB_DICT["job_id"]
    job = backend_for_retrieve_test.retrieve_job(job_id)
    assert job.job_id() == job_id


@pytest.mark.asyncio
async def test_backend_retrieve_job_async(backend_for_retrieve_test):
    job_id = _JOB_DICT["job_id"]
    job = await backend_for_retrieve_test.retrieve_job_async(job_id)
    assert job.job_id() == job_id


def test_job_base_interface(provider):
    backend = provider.get_backend(QCticQasmSimulator.NAME)
    job_id = str(uuid.uuid4())
    qobj = fake_qobj()
    job = QCticJob(backend, job_id, qobj)

    assert job.backend() == backend
    assert job.job_id() == job_id


@pytest.fixture
def job_for_status_test(job, httpserver):
    status = JobStatus.QUEUED.name
    url = "/jobs/{}".format(job.job_id())

    httpserver.expect_request(url).respond_with_json({
        "job_id": job.job_id(),
        "qobj": job.qobj().to_dict(),
        "date_submit": datetime.datetime.now().isoformat(),
        "status": status
    })

    return job


def test_job_status(job_for_status_test):
    assert job_for_status_test.status() == _STATUS


@pytest.mark.asyncio
async def test_job_status_async(job_for_status_test):
    the_status = await job_for_status_test.status_async()
    assert the_status == _STATUS


@pytest.fixture
def job_for_submit_test(job, httpserver):
    def handler(req):
        data = json.loads(req.data)
        assert data["job_id"] == job.job_id()
        assert data["qobj"] == job.qobj().to_dict()
        return Response()

    httpserver.expect_request("/jobs").respond_with_handler(handler)

    return job


def test_job_submit(job_for_submit_test):
    job_for_submit_test.submit()


@pytest.mark.asyncio
async def test_job_submit_async(job_for_submit_test):
    await job_for_submit_test.submit_async()


@pytest.fixture
def job_for_cancel_test(job, httpserver):
    def handler(req):
        data = json.loads(req.data)
        assert data["status"] == JobStatus.CANCELLED.name
        return Response()

    url = "/jobs/{}".format(job.job_id())
    httpserver.expect_request(url).respond_with_handler(handler)

    return job


def test_job_cancel(job_for_cancel_test):
    job_for_cancel_test.cancel()


@pytest.mark.asyncio
async def test_job_cancel_async(job_for_cancel_test):
    await job_for_cancel_test.cancel_async()


def test_job_result(job, httpserver):
    circuits, _run_config, _user_qobj_header = disassemble(job.qobj())
    simulator = Aer.get_backend("qasm_simulator")
    simulator_job = execute(circuits[0], simulator)
    simulator_result = simulator_job.result()
    url = "/jobs/{}".format(job.job_id())
    now = datetime.datetime.now()

    httpserver.expect_request(url).respond_with_json({
        "job_id": job.job_id(),
        "qobj": job.qobj().to_dict(),
        "date_submit": now.isoformat(),
        "date_start": (now + datetime.timedelta(1)).isoformat(),
        "date_end": (now + datetime.timedelta(2)).isoformat(),
        "status": JobStatus.DONE.name,
        "result": simulator_result.to_dict()
    })

    result = job.result()

    assert result.to_dict() == simulator_result.to_dict()


def test_job_fetch_remote(job, httpserver):
    job_dict = {
        "job_id": job.job_id(),
        "qobj": job.qobj().to_dict(),
        "date_submit": datetime.datetime.now().isoformat(),
        "status": JobStatus.INITIALIZING.name
    }

    count = [0]

    def handler(req):
        count[0] += 1
        return Response(json.dumps(job_dict))

    url = "/jobs/{}".format(job.job_id())
    httpserver.expect_request(url).respond_with_handler(handler)

    assert count[0] == 0
    assert job.status() == job_dict["status"]
    assert count[0] == 1
    assert job.status() == job_dict["status"]
    assert count[0] == 1
    assert job.status(fetch=True) == job_dict["status"]
    assert count[0] == 2



import asyncio
import collections
import logging
import os
import pprint
import random
import time
import warnings
from functools import wraps

from qiskit.providers import BaseJob, JobError, JobStatus
from qiskit.result import Result

from qctic.utils import wait_result, wait_result_async

_PARAM_TIMEOUT = "timeout"
_PARAM_WAIT = "wait"
_ENV_SIM_RATIO = "QCTIC_SIMULATION_RATIO_WARNING"

_logger = logging.getLogger(__name__)


def _fetch_job_sync(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        fetch = kwargs.pop("fetch", False)

        if not args[0]._remote_job or fetch:
            args[0].fetch_remote()

        return func(*args, **kwargs)

    return wrapper


def _fetch_job_async(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        fetch = kwargs.pop("fetch", False)

        if not args[0]._remote_job or fetch:
            await args[0].fetch_remote_async()

        return await func(*args, **kwargs)

    return wrapper


async def _dummy_submit_task():
    _logger.debug((
        "Waiting on dummy submit task: "
        "Was job.submit() in backend.run() synchronous?"
    ))


WaitResultParams = collections.namedtuple(
    "WaitResultParams",
    ["timeout", "wait"])


class SimulationRatioWarning(Warning):
    """Warning raised when the simulation ratio of any given job is too low. 
    This usually means that it would be more optimal for the experiments to be executed 
    locally to avoid paying the cost of network latencies and worker initialization."""


class QCticJob(BaseJob):
    """A Qiskit job that is executed remotely in the CTIC Erwin simulation platform."""

    def __init__(self, backend, job_id, qobj, run_params=None, remote_job=None):
        """Constructor.

        Args:
            backend (QCticQasmSimulator): Backend that contains this job.
            job_id (str): ID of the job.
            qobj (Qobj): Qobj that contains the experiment.
            run_params (dict): Optional dict of arbitrary arguments passed
                to the ``run`` method in the remote simulation platform.
        """

        super().__init__(backend, job_id)
        self._qobj = qobj
        self._remote_job = remote_job
        self._fetch_time = None
        self._run_params = run_params
        self._submit_task = None
        self._raised_warning = False

    @property
    def api(self):
        """QCticAPI: The API instance of this job's backend."""

        return self.backend().api

    def qobj(self):
        """Return the Qobj submitted for this job.

        Returns:
            Qobj: The Qobj submitted for this job.
        """

        return self._qobj

    @property
    def run_params(self):
        """dict: Arguments passed to the ``run`` method in the remote simulation platform."""

        return self._run_params

    @property
    def remote_job(self):
        """dict: Serialized version of this job as represented in the remote API."""

        return self._remote_job

    @property
    def time_submit(self):
        """datetime.timedelta: Time spent by the job from submission to the 
        moment it entered the queue. This includes the API network latency."""

        date_submit = self.remote_job.get("date_submit")
        date_queue = self.remote_job.get("date_queue")

        if not date_submit or not date_queue:
            return None

        return date_queue - date_submit

    @property
    def time_enqueued(self):
        """datetime.timedelta: Time spent by the job in the queue waiting to be executed."""

        date_queue = self.remote_job.get("date_queue")
        date_start = self.remote_job.get("date_start")

        if not date_queue or not date_start:
            return None

        return date_start - date_queue

    @property
    def time_running(self):
        """datetime.timedelta: Time spent by the job in an active worker 
        including the environment initialization and simulation."""

        date_start = self.remote_job.get("date_start")
        date_end = self.remote_job.get("date_end")

        if not date_start or not date_end:
            return None

        return date_end - date_start

    @property
    def time_simulation(self):
        """datetime.timedelta: Time spent by the job in the actual simulation."""

        date_ex_start = self.remote_job.get("date_execute_start")
        date_ex_end = self.remote_job.get("date_execute_end")

        if not date_ex_start or not date_ex_end:
            return None

        return date_ex_end - date_ex_start

    @property
    def time_total(self):
        """datetime.timedelta: Total time spent by the job from submission 
        to the moment the final simulation results are persisted."""

        date_submit = self.remote_job.get("date_submit")
        date_end = self.remote_job.get("date_end")

        if not date_submit or not date_end:
            return None

        return date_end - date_submit

    def get_simulation_ratio(self, queue=True):
        """Returns the ratio between the time spent in the actual simulation and the total time 
        including network latency, environment initialization and (optional) queue latency.

        Args:
            queue (bool): (Optional) Take the queue latency into account.

        Returns:
            float: The simulation ratio.
        """

        if not self.time_simulation or not self.time_total:
            return None

        if not queue and not self.time_enqueued:
            return None

        delta_total = self.time_total if queue else self.time_total - self.time_enqueued

        return round(self.time_simulation.total_seconds() / delta_total.total_seconds(), 3)

    @property
    def submit_task(self):
        return self._submit_task if self._submit_task else _dummy_submit_task()

    @submit_task.setter
    def submit_task(self, val):
        if self._submit_task:
            raise ValueError("The submit task has already been defined")

        self._submit_task = val

    def _log_fetch(self):
        _logger.debug("Fetching job: %s", self.job_id())

    def fetch_remote(self):
        self._log_fetch()
        self._remote_job = self.api.get_job_sync(self.job_id())
        self._fetch_time = time.time()

    async def fetch_remote_async(self):
        self._log_fetch()
        self._remote_job = await self.api.get_job(self.job_id())
        self._fetch_time = time.time()

    def _raise_simulation_warning(self):
        limit = os.getenv(_ENV_SIM_RATIO, None)

        if self._raised_warning or limit is None:
            return

        ratio = self.get_simulation_ratio(queue=False)

        if ratio is None:
            return

        try:
            limit = float(limit)
        except Exception as ex:
            _logger.warning("Invalid sim. ratio value (%s): %s", limit, ex)
            return

        if ratio < limit:
            warn_msg = (
                "The simulation time ratio of this job ({}) "
                "appears to be low (<{}). This is without taking "
                "into account wait times in the queue. Please consider "
                "running the simulation locally to optimize wait times."
            ).format(ratio, limit)

            _logger.warning(warn_msg)
            warnings.warn(warn_msg, SimulationRatioWarning)
            self._raised_warning = True

    def _result(self):
        error = self._remote_job.get("error", None)

        if error is not None:
            raise JobError(error)

        result_dict = self._remote_job.get("result", None)
        result = Result.from_dict(result_dict) if result_dict else None

        if result and not result.success:
            status_msg = pprint.pformat([r.status for r in result.results])
            _logger.warning("Some experiments failed:\n%s", status_msg)

        if result and result.success:
            self._raise_simulation_warning()

        return result

    def _status(self):
        return self._remote_job.get("status", None)

    def _get_wait_params(self, *args, **kwargs):
        if _PARAM_TIMEOUT not in kwargs or _PARAM_WAIT not in kwargs:
            return None

        wait = int(kwargs[_PARAM_WAIT])
        timeout = kwargs[_PARAM_TIMEOUT]

        if timeout is not None:
            timeout = int(timeout)

        _logger.debug(
            "Should wait for result (timeout=%s, wait=%s)",
            timeout, wait)

        return WaitResultParams(timeout=timeout, wait=wait)

    @_fetch_job_sync
    def status(self):
        """Return the status of the job, one of the names of the ``JobStatus`` enum.

        Args:
            fetch (bool): (Optional) Fetch the status from the remote API.

        Returns:
            JobStatus: The name of a member of the JobStatus enum.
        """

        return self._status()

    def submit(self):
        """Submit the job to the backend for execution."""

        self.api.post_job_sync(self)

    def cancel(self):
        """Attempt to cancel the job."""

        self.api.cancel_job_sync(self.job_id())

    @_fetch_job_sync
    def result(self, *args, **kwargs):
        """Return the results of the job.

        Args:
            fetch (bool): (Optional) Fetch the result from the remote API.

        Returns:
            Result: The results.

        Raises:
            JobError: If there was an error running the experiment on the remote simulator.
        """

        wait_params = self._get_wait_params(*args, **kwargs)

        if wait_params:
            return wait_result(
                self,
                base_sleep=wait_params.wait,
                timeout=wait_params.timeout,
                sleep_growth=1.0)

        return self._result()

    @_fetch_job_async
    async def status_async(self):
        return self._status()

    async def submit_async(self):
        await self.api.post_job(self)

    async def cancel_async(self):
        await self.api.cancel_job(self.job_id())

    @_fetch_job_async
    async def result_async(self, *args, **kwargs):
        wait_params = self._get_wait_params(*args, **kwargs)

        if wait_params:
            return await wait_result_async(
                self,
                base_sleep=wait_params.wait,
                timeout=wait_params.timeout,
                sleep_growth=1.0)

        return self._result()



import asyncio
import datetime
import logging
import re
from functools import wraps
from textwrap import shorten
from urllib.parse import urljoin

from qiskit.providers import JobStatus
from tornado.escape import json_decode, json_encode
from tornado.httpclient import (AsyncHTTPClient, HTTPClient, HTTPError,
                                HTTPRequest)
from tornado.httputil import url_concat

from qctic.schemas import BackendStatusSchema, GetJobsQuerySchema, JobSchema

_SCHEME_REGEX = r"^.+://.*"
_SCHEME_DEFAULT = "http"
_BODY_WIDTH = 1024

_logger = logging.getLogger(__name__)


def _format_req(req):
    base = "{} {}".format(req.method, req.url)
    body = shorten(req.body.decode(), width=_BODY_WIDTH) if req.body else None
    return "{}\n{}".format(base, body) if body else base


def _assert_synchronous(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if args[0].loop.is_running():
            raise RuntimeError((
                "Cannot use sync methods if the asyncio loop is running. "
                "Please use the async interface instead."
            ))

        return func(*args, **kwargs)

    return wrapper


class QCticAPIError(Exception):
    pass


class QCticAPI:
    """Class that represents the Erwin simulation API interface."""

    DEFAULT_HOST = "http://localhost"

    def __init__(self, host=DEFAULT_HOST):
        """Constructor.

        Args:
            host (str): (Optional) URL of the API server (``http://localhost`` by default).
        """

        if not re.match(_SCHEME_REGEX, host):
            _logger.warning(
                "Host '%s' has no scheme: Using '%s' by default",
                host, _SCHEME_DEFAULT)

            host = "{}://{}".format(_SCHEME_DEFAULT, host)

        self._req_defaults = {"headers": {"Content-Type": "application/json"}}
        self._client = None
        self._host = host
        self._loop = None

        if not self._host.endswith("/"):
            self._host = f"{self._host}/"

    @property
    def loop(self):
        """asyncio.AbstractEventLoop: The event loop used by the sync methods of this class."""

        if self._loop:
            return self._loop

        try:
            return asyncio.get_running_loop()
        except (RuntimeError, AttributeError):
            pass

        try:
            return asyncio.get_event_loop()
        except RuntimeError:
            # It seems we are not in the main thread,
            # thus the loop was not created implicitly.
            asyncio.set_event_loop(asyncio.new_event_loop())

        return asyncio.get_event_loop()

    @loop.setter
    def loop(self, val):
        self._loop = val

    @property
    def host(self):
        return self._host

    @property
    def client(self):
        """tornado.httpclient.AsyncHTTPClient: HTTP client."""

        if not self._client:
            self._client = AsyncHTTPClient()

        if not self.has_auth:
            raise Exception("Must authenticate first")

        return self._client

    @property
    def has_auth(self):
        """bool: True if the authentication arguments are defined."""

        token = self._req_defaults.get("headers", {}).get("Authorization")
        user = self._req_defaults.get("auth_username")
        passwd = self._req_defaults.get("auth_password")

        return token or (user and passwd)

    def _build_url(self, part):
        part = part if not part.startswith("/") else part[1:]
        return urljoin(self.host, part)

    def _clean_auth(self):
        self._req_defaults.pop("auth_username", None)
        self._req_defaults.pop("auth_password", None)
        self._req_defaults.get("headers", {}).pop("Authorization", None)

    def auth_token(self, token):
        """Update the HTTP client configuration to use *Bearer* authorization.

        Clears any authorization data that was previously defined.

        Args:
            token (str): API token.
        """

        self._clean_auth()

        self._req_defaults["headers"] = self._req_defaults.get("headers", {})

        self._req_defaults["headers"].update({
            "Authorization": "Bearer {}".format(token)
        })

    def auth_basic(self, user, passwd):
        """Update the HTTP client configuration to use *Basic* authorization.

        Clears any authorization data that was previously defined.

        Args:
            user (str): Username.
            passwd (str): Password.
        """

        self._clean_auth()

        self._req_defaults.update({
            "auth_username": user,
            "auth_password": passwd
        })

    def _request(self, **kwargs):
        request_params = {**self._req_defaults}
        request_params.update(kwargs)
        return HTTPRequest(**request_params)

    async def _fetch(self, *args, **kwargs):
        try:
            return await self.client.fetch(*args, **kwargs)
        except HTTPError as ex:
            err = json_decode(ex.response.body)
            _logger.warning("Error on API request: %s", err)
            err_msg = "[{}] {}".format(err.get("name"), err.get("description"))
            raise QCticAPIError(err_msg)

    async def get_job(self, job_id):
        """Retrieves a single Job from the API.

        Args:
            job_id (str): Job ID.

        Returns:
            dict: Dict that conforms to the ``JobSchema`` schema.
        """

        req = self._request(
            url=self._build_url("/jobs/{}".format(job_id)),
            method="GET")

        _logger.debug(_format_req(req))
        res = await self._fetch(req)

        return JobSchema().load(json_decode(res.body))

    async def get_jobs(self, **kwargs):
        """Retrieves a set of Jobs from the API.

        Args:
            limit (int): (Optional) Maximum number of Jobs that should be returned.
            skip (int): (Optional) Skip this number of Jobs before adding them to the result set 
                (useful for pagination). Jobs are sorted by date_submit DESC by default.
            status (list(str)): (Optional) Set of status used as job filter.
            date_start (datetime): (Optional) Lower threshold for ``date_submit``.
            date_end (datetime): (Optional) Upper threshold for ``date_submit``.

        Returns:
            list(dict): List of dicts that conform to the ``JobSchema`` schema.
        """

        params = GetJobsQuerySchema().dump(kwargs)
        status = params.pop("status", [])
        url = url_concat(self._build_url("/jobs"), params)

        for item in status:
            url = url_concat(url, {"status": item})

        req = self._request(url=url, method="GET")

        _logger.debug(_format_req(req))
        res = await self._fetch(req)

        return [JobSchema().load(item) for item in json_decode(res.body)]

    async def get_backend_status(self):
        """Retrieves the current status of the simulation platform.

        Returns:
            dict: Dict that conforms to the ``BackendStatusSchema`` schema.
        """

        req = self._request(
            url=self._build_url("/status"),
            method="GET")

        _logger.debug(_format_req(req))
        res = await self._fetch(req)

        return BackendStatusSchema().load(json_decode(res.body))

    async def post_job(self, job):
        """Creates a new job in the simulation platform.

        Args:
            job (QCticJob): The job to be created.
        """

        job_init = {
            "qobj": job.qobj().to_dict(),
            "job_id": job.job_id(),
            "status": JobStatus.INITIALIZING.name,
            "date_submit": datetime.datetime.now(datetime.timezone.utc)
        }

        if job.run_params and len(job.run_params) > 0:
            job_init.update({"run_params": job.run_params})

        body = json_encode(JobSchema().dump(job_init))

        req = self._request(
            url=self._build_url("/jobs"),
            method="POST",
            body=body)

        _logger.debug(_format_req(req))
        await self._fetch(req)

    async def cancel_job(self, job_id):
        """Attempts to cancel a job that is currently active.

        Args:
            job_id (str): Job ID.
        """

        body = json_encode({"status": JobStatus.CANCELLED.name})

        req = self._request(
            url=self._build_url("/jobs/{}".format(job_id)),
            method="PUT",
            body=body)

        _logger.debug(_format_req(req))
        await self._fetch(req)

    @_assert_synchronous
    def get_job_sync(self, job_id):
        """Synchronous version of the ``get_job`` method."""

        return self.loop.run_until_complete(self.get_job(job_id))

    @_assert_synchronous
    def get_jobs_sync(self, **kwargs):
        """Synchronous version of the ``get_jobs`` method."""

        return self.loop.run_until_complete(self.get_jobs(**kwargs))

    @_assert_synchronous
    def get_backend_status_sync(self):
        """Synchronous version of the ``get_backend_status`` method."""

        return self.loop.run_until_complete(self.get_backend_status())

    @_assert_synchronous
    def post_job_sync(self, job):
        """Synchronous version of the ``post_job`` method."""

        return self.loop.run_until_complete(self.post_job(job))

    @_assert_synchronous
    def cancel_job_sync(self, job_id):
        """Synchronous version of the ``cancel_job`` method."""

        return self.loop.run_until_complete(self.cancel_job(job_id))



import argparse
import asyncio
import logging
import os
import pprint
import random

from qiskit import QuantumCircuit, execute

from qctic.api import QCticAPI
from qctic.backend import (QCticQasmSimulator, QCticStatevectorSimulator,
                           QCticUnitarySimulator)
from qctic.provider import QCticProvider
from qctic.utils import wait_result_async

_MIN_QUBITS = 2
_PROB_H = 0.5

_logger = logging.getLogger("qctic")


def init_logging():
    try:
        import coloredlogs
        coloredlogs.install(level=logging.DEBUG, logger=_logger)
    except ImportError:
        _logger.setLevel(logging.DEBUG)
        logging.basicConfig()


# pylint: disable=no-member
def random_circuit(max_qubits, unitary=False):
    max_qubits = max_qubits if max_qubits >= _MIN_QUBITS else _MIN_QUBITS
    qubits = random.randint(_MIN_QUBITS, max_qubits)
    circuit = QuantumCircuit(qubits, qubits)

    h_gates = [
        idx for idx in range(qubits)
        if random.random() <= _PROB_H
    ]

    if len(h_gates) == 0:
        h_gates.append(0)

    [circuit.h(idx) for idx in h_gates]

    if not unitary:
        measure_args = (list(range(qubits)),) * 2
        circuit.measure(*measure_args)

    return circuit


def build_result_msg_qasm(result, circuit):
    return "## Counts:\n{}\n## Memory:\n{}".format(
        pprint.pformat(result.get_counts(circuit)),
        pprint.pformat(result.get_memory(circuit)))


def build_result_msg_unitary(result, circuit):
    return "## Unitary:\n{}".format(
        pprint.pformat(result.get_unitary(circuit)))


def build_result_msg_statevector(result, circuit):
    return "## Statevector:\n{}".format(
        pprint.pformat(result.get_statevector(circuit)))


def log_result(result, circuits, backend_name):
    msg_funcs = {
        QCticQasmSimulator.NAME: build_result_msg_qasm,
        QCticUnitarySimulator.NAME: build_result_msg_unitary,
        QCticStatevectorSimulator.NAME: build_result_msg_statevector
    }

    msg_func = msg_funcs[backend_name]

    for idx, circ in enumerate(circuits):
        _logger.info("Circuit #%s results:\n%s", idx, msg_func(result, circ))


async def execute_jobs(qubits, backend_name, num_circuits, shots, gpu):
    qctic_provider = QCticProvider.from_env()
    qctic_backend = qctic_provider.get_backend(backend_name)

    is_unitary = backend_name == QCticUnitarySimulator.NAME

    circuits = [
        random_circuit(max_qubits=qubits, unitary=is_unitary)
        for _idx in range(num_circuits)
    ]

    for idx, circ in enumerate(circuits):
        _logger.info("Circuit #%s:\n%s", idx, circ.draw())

    backend_options = {}

    if gpu and backend_name == QCticQasmSimulator.NAME:
        backend_options.update({"method": "statevector_gpu"})
    elif gpu and backend_name != QCticQasmSimulator.NAME:
        _logger.warning("Ignoring GPU flag")

    memory = backend_name == QCticQasmSimulator.NAME

    qctic_job = execute(
        circuits,
        qctic_backend,
        shots=shots,
        backend_options=backend_options,
        memory=memory)

    await qctic_job.submit_task
    result = await wait_result_async(qctic_job)

    if not result.success:
        _logger.error("Some of the experiments failed")

        for idx, exp_res in enumerate(result.results):
            _logger.warning("Circuit #%s status: %s", idx, exp_res.status)

        return

    log_result(result=result, circuits=circuits, backend_name=backend_name)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--circuits", type=int, default=1)
    parser.add_argument("--shots", type=int, default=10)
    parser.add_argument("--qubits", type=int, default=4)
    parser.add_argument("--gpu", action="store_true")

    backend_choices = [
        QCticQasmSimulator.NAME,
        QCticStatevectorSimulator.NAME,
        QCticUnitarySimulator.NAME
    ]

    parser.add_argument(
        "--backend",
        default=QCticQasmSimulator.NAME,
        choices=backend_choices)

    return parser.parse_args()


def main():
    init_logging()
    loop = asyncio.get_event_loop()
    args = parse_args()

    _logger.debug("Arguments: %s", args)

    def stop_cb(fut):
        _logger.debug("Main task: %s", fut)
        loop.stop()

    try:
        main_task = asyncio.ensure_future(execute_jobs(
            qubits=args.qubits,
            backend_name=args.backend,
            num_circuits=args.circuits,
            shots=args.shots,
            gpu=args.gpu))

        main_task.add_done_callback(stop_cb)
        loop.run_forever()
    except:
        _logger.error("Error", exc_info=True)
    finally:
        _logger.debug("Closing loop")
        loop.close()


if __name__ == "__main__":
    main()
